from flask import Flask, send_from_directory, request
from flask_restful import Api, Resource, reqparse
import json
import numpy as np
import datetime
import csv

from dlw import DLWSubject

STATICS_LOCATION = 'dist'

app = Flask(__name__, static_url_path='', static_folder=STATICS_LOCATION)
api = Api(app)

CALCULATED_RESULTS = None  # type: DLWSubject


@app.route('/calculate', methods=['POST'])
def calculate_from_inputs():
    input_data = json.loads(request.get_data().decode('utf-8'))
    datetimes = [datetime.datetime(l[0], l[1], l[2], l[3], l[4]) for l in input_data['datetimes']]
    global CALCULATED_RESULTS
    CALCULATED_RESULTS = DLWSubject(d_meas=np.asarray(input_data['d_meas'], dtype=float),
                                    o18_meas=np.asarray(input_data['o18_meas'], dtype=float),
                                    sample_datetimes=np.asarray(datetimes),
                                    dose_weights=np.asarray(input_data['dose_weights'], dtype=float),
                                    mixed_dose=input_data['mixed_dose'],
                                    dose_enrichments=np.asarray(input_data['dose_enrichments'], dtype=float),
                                    subject_weights=np.asarray(input_data['subject_weights'], dtype=float),
                                    subject_id=input_data['subject_id'],
                                    in_permil=input_data['in_permil'])

    def sort_calculated_results(results):
        return {
            "rco2_ee_int": {
                "rco2_mol_day": ["rCO2 (mol/day)", round(results['co2_int_mol_day'], 2)],
                "rco2_l_hr": ["rCO2 (L/day)", round(results['co2_int_L_day'], 1)],
                "ee_kcal_day": ["EE (kcal/day)", round(results['tee_int_kcal_day'], 1)],
                "ee_mj_day": ["EE (MJ/day)", round(results['tee_int_mj_day'], 2)]
            },
            "rco2_ee_plat": {
                "rco2_mol_day": ["rCO2 (mol/day)", round(results['co2_plat_mol_day'], 2)],
                "rco2_l_hr": ["rCO2 (L/day)", round(results['co2_plat_L_day'], 1)],
                "ee_kcal_day": ["EE (kcal/day)", round(results['tee_plat_kcal_day'], 1)],
                "ee_mj_day": ["EE (MJ/day)", round(results['tee_plat_mj_day'], 2)]
            }
        }

    return json.dumps({
        "results": {
            "calculations": {
                "ndp_kg": ["NdP (kg)", round(CALCULATED_RESULTS.nd['adj_plat_avg_kg'], 1)],
                "kd_hr": ["kd/hour", round(CALCULATED_RESULTS.kd_per_hr, 6)],
                "nop_kg": ["NoP (kg)", round(CALCULATED_RESULTS.no['adj_plat_avg_kg'], 1)],
                "ko_hr": ["ko/hour", round(CALCULATED_RESULTS.ko_per_hr, 6)],
                "body_water_avg_kg": ["Total Body Water Average (kg)",
                                      round(CALCULATED_RESULTS.total_body_water_ave_kg, 1)],
                "fat_free_mass_kg": ["Fat Free Mass (kg)", round(CALCULATED_RESULTS.fat_free_mass_kg, 1)],
                "fat_mass_kg": ["Fat Mass (kg)", round(CALCULATED_RESULTS.fat_mass_kg, 1)],
                "body_fat_percentage": ["Body Fat Percentage", round(CALCULATED_RESULTS.body_fat_percent, 1)]
            },
            "error_flags": {
                "plateau_2h": ["2H plateau (<5%)", round(CALCULATED_RESULTS.d_ratio_percent, 2)],
                "plateau_180": ["18O Plateau (<5%)", round(CALCULATED_RESULTS.o18_ratio_percent, 2)],
                "ds_ratio": ["DS ratio (1.000 - 1.070)", round(CALCULATED_RESULTS.dil_space_ratio, 4)],
                "ee": ["EE, Schoeller (PD4-ED4 vs. PD5-ED5, <10%)", round(CALCULATED_RESULTS.ee_check, 4)],
                "ko_kd": ["Ko/kd (1.1 - 1.7)", round(CALCULATED_RESULTS.ko_kd_ratio, 4)]
            },
            "schoeller": sort_calculated_results(CALCULATED_RESULTS.schoeller),
            "racette": sort_calculated_results(CALCULATED_RESULTS.racette),
            "speakman": sort_calculated_results(CALCULATED_RESULTS.speakman)
        }
    })


@app.route('/export', methods=['POST'])
def export_to_csv():
    input_data = json.loads(request.get_data().decode('utf-8'))
    csv_filename = input_data['filename']
    if csv_filename.endswith('.csv'):
        saved_file = CALCULATED_RESULTS.save_results_csv(csv_filename)
    else:
        saved_file = CALCULATED_RESULTS.save_results_csv(csv_filename + '.csv')
    return json.dumps({"saved_file": saved_file})


@app.route('/load', methods=['POST'])
def load_csv():
    file = request.get_data().decode('utf-8')
    rows = file.split('\n')
    reader = csv.DictReader(rows)
    for row in reader:
        # there should only be one, TODO batched processing
        return json.dumps({'results': row, 'error': False})


@app.route('/')
def root():
    return send_from_directory(STATICS_LOCATION, 'index.html')


def run_app():
    app.run(debug=False)


if __name__ == '__main__':
    app.run(debug=True)



import numpy as np
from datetime import timedelta
import os.path

DEUTERIUM = "deuterium"
OXYGEN = "oxygen"

D_VSMOW_RATIO = 0.00015576
O18_VSMOW_RATIO = 0.0020052
STANDARD_WATER_MOL_MASS = 18.10106 / 1000  # kg
PPM_TO_RATIO = 1 / 1000000
POP_DIL_SPACE_D = 1.041
POP_DIL_SPACE_O = 1.007
STD_POP_AVG_RDIL = 1.02
FAT_FREE_MASS_FACTOR = 0.73
HOURS_PER_DAY = 24
LITERS_PER_MOL = 22.414
WEIR_CONSTANT = 5.7425
MJ_PER_KCAL = 4.184 / 1000

D_PLATEAU_LIMIT = 5.0
O18_PLATEAU_LIMIT = 5.0
DIL_SPACE_RATIO_LOW_LIMIT = 1.00
DIL_SPACE_RATIO_HIGH_LIMIT = 1.07
KO_KD_RATIO_LOW_LIMIT = 1.1
KO_KD_RATIO_HIGH_LIMIT = 1.7


class DLWSubject:
    """Class for performing Doubly Labeled Water calculations
       Attributes:
           d_meas (np.array): deuterium values as measured (can be in permil or ppm)
           o18_meas (np.array): oxygen 18 values as measured (can be in permil or ppm)
           d_deltas (np.array): deuterium delta values of subject samples
           o18_deltas (np.array): oxygen 18 delta values of subject samples
           sample_datetimes ([datetime]): dates and times of sample collections and dose
                                        should be in order of background, dose, PD4, PD5, ED4, ED5
           dose_weights ([float]): weights in g of doses administered, deuterium first, 18O second
           mixed_dose ([bool]): boolean indicating whether doses are mixed together or separate
           mol_masses ([float]): molecular masses in g/mol of doses administered, deuterium first, 18O second
           dose_enrichments ([float]): dose enrichments in ratio of doses administered, deuterium first, 18O second
           subject_weights ([float]): initial and final weights of the subject in kg
           subject_id ([string]): string identifier for the data
           pop_avg_rdil ([float]): population average dilution space ratio, if provided
           in_permil ([bool]): True if measured d and O18 are in permil, false if they are in ppm
           d_ratios (np.array): deuterium ratios of subject samples
           o18_ratios (np.array): 18O ratios of subject samples
           kd_per_hr (float): deuterium turnover rate in 1/hr
           ko_per_hr (float): oxygen turnover rate in 1/hr
           ko_kd_ratio (float): ratio of oxygen and deuterium turnover rates
           nd (dict): dictionary containing all the calculated values of the deuterium dilution space
                plat_a_mol (float): calculated by the plateau method using the a sample in mol
                plat_b_mol (float): calculated by the plateau method using the b sample in mol
                plat_avg_mol (float): the average of the a and b plateau dilution spaces in mol
                int_a_mol (float): calculated by the intercept method using the a sample in mol
                int_b_mol (float): calculated by the intercept method using the b sample in mol
                int_avg_mol (float): the average of the a and b intercept dilution spaces in mol
                adj_plat_avg_mol (float): average plateau dilution space adjusted for the subject weight change over
                           the sampling period, in mol
                adj_int_avg_mol (float): average intercept dilution space adjusted for the subject weight change over
                           the sampling period, in mol
                adj_plat_avg_kg (float): average plateau dilution space adjusted for the subject weight change over
                           the sampling period, in kg
                adj_int_avg_kg (float): average intercept dilution space adjusted for the subject weight change over
                           the sampling period, in kg
           no (dict): dictionary containing all the calculated values of the oxygen 18 dilution space
                plat_a_mol (float): calculated by the plateau method using the a sample in mol
                plat_b_mol (float): calculated by the plateau method using the b sample in mol
                plat_avg_mol (float): the average of the a and b plateau dilution spaces in mol
                int_a_mol (float): calculated by the intercept method using the a sample in mol
                int_b_mol (float): calculated by the intercept method using the b sample in mol
                int_avg_mol (float): the average of the a and b intercept dilution spaces in mol
                adj_plat_avg_mol (float): average plateau dilution space adjusted for the subject weight change over
                           the sampling period, in mol
                adj_int_avg_mol (float): average intercept dilution space adjusted for the subject weight change over
                           the sampling period, in mol
                adj_plat_avg_kg (float): average plateau dilution space adjusted for the subject weight change over
                           the sampling period, in kg
                adj_int_avg_kg (float): average intercept dilution space adjusted for the subject weight change over
                           the sampling period, in kg
           dil_space_ratio (float): ratio of the 4 hr plateau deuterium and 18O dilution spaces
           total_body_water_d_kg (float): total body water calculated from the average D plateau dilution space, in kg
           total_body_water_o_kg (float): total body water calculated from the average 18O plateau dilution space, in kg
           total_body_water_ave_kg (float): average of total_body_water_d_kg and total_body_water_o_kg, in kg
           fat_free_mass_kg (float): fat free mass of the subject, in kg
           fat_mass_kg (float): fat mass of the subject, in kg
           body_fat_percent (float): body fat percent of the subject, in percent
           schoeller (dict): dictionary containing all the values calculated using the equation of Schoeller (equation
                            A6, 1986 as updated in 1988)
                co2_int (float): CO2 production rate in mol/hr using the weight adjusted, average, intercept dilution spaces
                co2_plat (float): CO2 production rate in mol/hr using the weight adjusted, average, plateau dilution spaces
                co2_int_mol_day (float): CO2 production rate in mol/day using the weight adjusted, average, intercept dilution spaces
                co2_int_L_hr (float): CO2 production rate in L/hr using the weight adjusted, average, intercept dilution spaces
                tee_int_kcal_day (float): Total energy expenditure calculated using the equation of Weir (1949) from co2
                            values calculated via Schoeller and the weight adjusted, average, intercept dilution spaces,
                            in kcal/day
                tee_plat_kcal_day (float): Total energy expenditure calculated using the equation of Weir (1949) from co2
                              values calculated via Schoeller and the weight adjusted, average, plateau dilution spaces,
                              in kcal/day
                tee_int_mj_day (float): Total energy expenditure calculated using the equation of Weir (1949) from
                            co2 values calculated via Schoeller and the weight adjusted, average, intercept dilution
                            spaces, in mj/day
                tee_plat_mj_day (float): Total energy expenditure calculated using the equation of Weir (1949) from
                            co2 values calculated via Schoeller and the weight adjusted, average, plateau dilution
                            spaces, in mj/day
            racette (dict): dictionary containing all the values calculated using the equation of Racette, eqn 1 (1994)
                co2_int (float): CO2 production rate in mol/hr using the weight adjusted, average, intercept dilution spaces
                co2_plat (float): CO2 production rate in mol/hr using the weight adjusted, average, plateau dilution spaces
                co2_int_mol_day (float): CO2 production rate in mol/day using the weight adjusted, average, intercept dilution spaces
                co2_int_L_hr (float): CO2 production rate in L/hr using the weight adjusted, average, intercept dilution spaces
                tee_int_kcal_day (float): Total energy expenditure calculated using the equation of Weir (1949) from co2
                            values calculated via Racette and the weight adjusted, average, intercept dilution spaces,
                            in kcal/day
                tee_plat_kcal_day (float): Total energy expenditure calculated using the equation of Weir (1949) from co2
                              values calculated via Racette and the weight adjusted, average, plateau dilution spaces,
                              in kcal/day
                tee_int_mj_day (float): Total energy expenditure calculated using the equation of Weir (1949) from
                            co2 values calculated via Racette and the weight adjusted, average, intercept dilution
                            spaces, in mj/day
                tee_plat_mj_day (float): Total energy expenditure calculated using the equation of Weir (1949) from
                            co2 values calculated via Racette and the weight adjusted, average, plateau dilution
                            spaces, in mj/day
            speakman (dict): dictionary containing all the values calculated using the equation of Speakman, eqn 17.41 (1997)
                co2_int (float): CO2 production rate in mol/hr using the weight adjusted, average, intercept dilution spaces
                co2_plat (float): CO2 production rate in mol/hr using the weight adjusted, average, plateau dilution spaces
                co2_int_mol_day (float): CO2 production rate in mol/day using the weight adjusted, average, intercept dilution spaces
                co2_int_L_hr (float): CO2 production rate in L/hr using the weight adjusted, average, intercept dilution spaces
                tee_int_kcal_day (float): Total energy expenditure calculated using the equation of Weir (1949) from co2
                            values calculated via Speakman and the weight adjusted, average, intercept dilution spaces,
                            in kcal/day
                tee_plat_kcal_day (float): Total energy expenditure calculated using the equation of Weir (1949) from co2
                              values calculated via Speakman and the weight adjusted, average, plateau dilution spaces,
                              in kcal/day
                tee_int_mj_day (float): Total energy expenditure calculated using the equation of Weir (1949) from
                            co2 values calculated via Speakman and the weight adjusted, average, intercept dilution
                            spaces, in mj/day
                tee_plat_mj_day (float): Total energy expenditure calculated using the equation of Weir (1949) from
                            co2 values calculated via Speakman and the weight adjusted, average, plateau dilution
                            spaces, in mj/day
           d_ratio_percent (float): Percent difference between the a and b delta measurements of deuterium
           o18_ratio_percent (float): Percent difference between the a and b delta measurements of 18O
           ee_check (float): Data quality check consisting of the percent difference between the TEE (in kcal/day)
                            calculated using the PD4/ED4 pair and the TEE calculated using the PD5/ED5 pair, both with
                            the plateau method.
    """

    def __init__(self, d_meas, o18_meas, sample_datetimes, dose_weights, mixed_dose, dose_enrichments,
                 subject_weights, subject_id, in_permil=True, pop_avg_rdil=None):
        """Constructor for the DLWSubject class
           :param d_meas (np.array): deuterium delta values of subject samples
           :param o18_meas (np.array): oxygen 18 delta values of subject samples
           :param sample_datetimes ([datetime]): dates and times of sample collections
           :param dose_weights ([float]): weights in g of doses administered, deuterium first, 18O second
           :param mixed_dose ([bool]): boolean indicating whether doses are mixed together or separate
           :param dose_enrichments ([float]): dose enrichments in ppm of doses administered, deuterium first, 18O second
           :param subject_weights ([float]): initial and final weights of the subject in kg
           :param subject_id ([string]): string identifier for the data
           :param in_permil ([bool]): True if measured d and O18 are in permil, false if they are in ppm
           :param pop_avg_rdil ([float]): population average dilution space to use in the calculations
        """
        if len(d_meas) == len(o18_meas) == len(sample_datetimes)-1 == 5:

            self.sample_datetimes = sample_datetimes
            self.dose_weights = dose_weights
            self.mixed_dose = mixed_dose
            self.dose_enrichments = np.array(dose_enrichments) / 1000000  # convert from ppm to ratio
            self.subject_weights = subject_weights
            self.subject_id = subject_id

            if pop_avg_rdil is None:
                self.pop_avg_rdil = STD_POP_AVG_RDIL
            else:
                self.pop_avg_rdil = pop_avg_rdil

            if (in_permil):
                self.d_deltas = d_meas
                self.o18_deltas = o18_meas
                self.d_ratios = self.d_deltas_to_ratios()
                self.o18_ratios = self.o18_deltas_to_ratios()
            else:
                self.d_ratios = d_meas * PPM_TO_RATIO
                self.o18_ratios = o18_meas * PPM_TO_RATIO
                self.d_deltas = self.d_ratios_to_deltas()
                self.o18_deltas = self.o18_ratios_to_deltas()

            self.kd_per_hr = self.average_turnover_2pt(self.d_ratios, self.sample_datetimes)
            self.ko_per_hr = self.average_turnover_2pt(self.o18_ratios, self.sample_datetimes)
            self.ko_kd_ratio = self.ko_per_hr / self.kd_per_hr

            self.mol_masses = self.calculate_mol_masses(self.dose_enrichments, self.mixed_dose)

            self.nd = self.calculate_various_nd()
            self.no = self.calculate_various_no()

            self.dil_space_ratio = self.nd['plat_a_mol'] / self.no['plat_a_mol']  # dilution space ratio err flag

            # self.rh2o = (self.nd['adj_plat_avg_kg'] * self.kd_per_hr * HOURS_PER_DAY) / STANDARD_WATER_MOL_MASS

            self.total_body_water_d_kg = self.nd['adj_plat_avg_kg'] / POP_DIL_SPACE_D
            self.total_body_water_o_kg = self.no['adj_plat_avg_kg'] / POP_DIL_SPACE_O
            self.total_body_water_ave_kg = (self.total_body_water_d_kg + self.total_body_water_o_kg) / 2
            # average total body water

            self.fat_free_mass_kg = self.total_body_water_ave_kg / FAT_FREE_MASS_FACTOR
            self.fat_mass_kg = self.subject_weights[0] - self.fat_free_mass_kg
            self.body_fat_percent = self.fat_mass_kg / self.subject_weights[0] * 100

            self.schoeller = self.calculate_schoeller()
            self.racette = self.calculate_racette()
            self.speakman = self.calculate_speakman()

            self.d_ratio_percent = self.percent_difference(self.d_ratios[1] - self.d_ratios[0],
                                                           self.d_ratios[2] - self.d_ratios[0])
            # err flag 2 h plateau < 5%
            self.o18_ratio_percent = self.percent_difference(self.o18_ratios[1] - self.o18_ratios[0],
                                                             self.o18_ratios[2] - self.o18_ratios[0])
            # err flag o18 plateau < 5%
            self.ee_check = self.ee_consistency_check()  # err flag # 4 pd4

        else:
            raise ValueError('Arrays not correct size')

    def d_deltas_to_ratios(self):
        """Convert deuterium delta values to ratios.
           :return: deuterium ratios
        """
        return ((self.d_deltas / 1000) + 1) * D_VSMOW_RATIO

    def o18_deltas_to_ratios(self):
        """Convert 18O delta values to ratios.
           :return: 18O ratios
        """
        return ((self.o18_deltas / 1000) + 1) * O18_VSMOW_RATIO

    def d_ratios_to_deltas(self):
        """Convert deuterium ratio values to delta.
           :return: deuterium deltas
        """
        return (self.d_ratios / D_VSMOW_RATIO - 1) * 1000

    def o18_ratios_to_deltas(self):
        """Convert 18O ratio values to deltas.
           :return: 18O deltas
        """
        return ((self.o18_ratios / O18_VSMOW_RATIO - 1) * 1000)

    @staticmethod
    def isotope_turnover_2pt(background, initratio, finalratio, elapsedhours):
        """Calculate an isotope turnover rate in 1/hr using the 2pt method
           :param background:  enrichment of the background urine measurement as a ratio
           :param initratio: enrichment of the initial urine measurement as a ratio
           :param finalratio: enrichment of the final urine measurement as a ratio
           :param elapsedhours: elapsed time in hours between the intial and final urine measurements
           :return: istope turnover rate in 1/hr
        """
        if (background < initratio and background < finalratio and finalratio < initratio):
            return (np.log(initratio - background) - np.log(finalratio - background)) / elapsedhours
        else:
            raise ValueError('Isotope ratios do not conform to pattern background < final < plateau')

    def average_turnover_2pt(self, ratios, sampledatetime):
        """Calculate the average isotope turnover rate in 1/hr using the 2pt method
           :param ratios: measured urine isotope ratios
           :param sampledatetime: time and date of urine collections
           :return: average isotope turnover rate in 1/hr
        """
        turnovers = np.zeros(4)

        elapsedhours = (timedelta.total_seconds(sampledatetime[4] - sampledatetime[2])) / 3600
        turnovers[0] = self.isotope_turnover_2pt(ratios[0], ratios[1], ratios[3], elapsedhours)

        elapsedhours = (timedelta.total_seconds(sampledatetime[5] - sampledatetime[2])) / 3600
        turnovers[1] = self.isotope_turnover_2pt(ratios[0], ratios[1], ratios[4], elapsedhours)

        elapsedhours = (timedelta.total_seconds(sampledatetime[4] - sampledatetime[3])) / 3600
        turnovers[2] = self.isotope_turnover_2pt(ratios[0], ratios[2], ratios[3], elapsedhours)

        elapsedhours = (timedelta.total_seconds(sampledatetime[5] - sampledatetime[3])) / 3600
        turnovers[3] = self.isotope_turnover_2pt(ratios[0], ratios[2], ratios[4], elapsedhours)
        return np.mean(turnovers)

    @staticmethod
    def calculate_mol_masses(dose_enrichments, mixed_dose):
        """ Calculate the molecular masses for the enriched dose waters
            :param: dose_enrichments ([float]):dose enrichments in ppm of doses administered, 2H first, 18O second
            :param: mixed_dose ([bool]): boolean indicating whether doses are mixed together or separate
            :return: array of floats of molecular masses, 2H first, 18O second.  Both numbers will be the same in
                        the case of a mixed dose
        """

        if mixed_dose:
            mol_mass = 2 * ((1 - dose_enrichments[0]) * 1 + dose_enrichments[0] * 2) + (
                    (1 - dose_enrichments[1]) * 16 + dose_enrichments[1] * 18)
            mol_masses = [mol_mass, mol_mass]
        else:
            mol_mass_d = 2 * ((1 - dose_enrichments[0]) * 1 + dose_enrichments[0] * 2) + (
                    (1 - O18_VSMOW_RATIO) * 16 + O18_VSMOW_RATIO * 18)
            mol_mass_o = 2 * ((1 - D_VSMOW_RATIO) * 1 + D_VSMOW_RATIO * 2) + (
                    (1 - dose_enrichments[1]) * 16 + dose_enrichments[1] * 18)
            mol_masses = [mol_mass_d, mol_mass_o]

        return mol_masses

    def calculate_various_nd(self):
        """Calculate the various dilution spaces for nd.
            :return: dict of dilution spaces for deuterium
        """

        nd = {}
        nd['plat_a_mol'] = self.dilution_space_plateau(self.dose_weights[0], self.mol_masses[0],
                                                         self.dose_enrichments[0], self.d_ratios[1], self.d_ratios[0])
        nd['plat_b_mol'] = self.dilution_space_plateau(self.dose_weights[0], self.mol_masses[0],
                                                         self.dose_enrichments[0], self.d_ratios[2], self.d_ratios[0])
        nd['plat_avg_mol'] = (nd['plat_a_mol'] + nd['plat_b_mol']) / 2

        dosetime = timedelta.total_seconds(self.sample_datetimes[2] - self.sample_datetimes[1]) / 3600
        nd['int_a_mol'] = self.dilution_space_intercept(self.dose_weights[0], self.mol_masses[0],
                                                          self.dose_enrichments[0], self.d_ratios[1], self.d_ratios[0],
                                                          self.kd_per_hr, dosetime)
        dosetime = timedelta.total_seconds(self.sample_datetimes[3] - self.sample_datetimes[1]) / 3600
        nd['int_b_mol'] = self.dilution_space_intercept(self.dose_weights[0], self.mol_masses[0],
                                                          self.dose_enrichments[0], self.d_ratios[2], self.d_ratios[0],
                                                          self.kd_per_hr, dosetime)
        nd['int_avg_mol'] = (nd['int_a_mol'] + nd['int_b_mol']) / 2

        nd['adj_plat_avg_mol'] = self.adj_dilution_space(nd['plat_avg_mol'], self.subject_weights)
        nd['adj_int_avg_mol'] = self.adj_dilution_space(nd['int_avg_mol'], self.subject_weights)
        nd['adj_plat_avg_kg'] = nd['adj_plat_avg_mol'] * STANDARD_WATER_MOL_MASS
        nd['adj_int_avg_kg'] = nd['adj_int_avg_mol'] * STANDARD_WATER_MOL_MASS
        return nd

    def calculate_various_no(self):
        """Calculate the various dilution spaces for nd.
            :return: dict of dilution spaces for deuterium
        """

        no = {}
        no['plat_a_mol'] = self.dilution_space_plateau(self.dose_weights[1], self.mol_masses[1],
                                                         self.dose_enrichments[1], self.o18_ratios[1],
                                                         self.o18_ratios[0])
        no['plat_b_mol'] = self.dilution_space_plateau(self.dose_weights[1], self.mol_masses[1],
                                                         self.dose_enrichments[1], self.o18_ratios[2],
                                                         self.o18_ratios[0])
        no['plat_avg_mol'] = (no['plat_a_mol'] + no['plat_b_mol']) / 2

        dosetime = timedelta.total_seconds(self.sample_datetimes[2] - self.sample_datetimes[1]) / 3600
        no['int_a_mol'] = self.dilution_space_intercept(self.dose_weights[1], self.mol_masses[1],
                                                          self.dose_enrichments[1], self.o18_ratios[1],
                                                          self.o18_ratios[0], self.ko_per_hr, dosetime)
        dosetime = timedelta.total_seconds(self.sample_datetimes[3] - self.sample_datetimes[1]) / 3600
        no['int_b_mol'] = self.dilution_space_intercept(self.dose_weights[1], self.mol_masses[1],
                                                          self.dose_enrichments[1], self.o18_ratios[2],
                                                          self.o18_ratios[0], self.ko_per_hr, dosetime)
        no['int_avg_mol'] = (no['int_a_mol'] + no['int_b_mol']) / 2

        no['adj_plat_avg_mol'] = self.adj_dilution_space(no['plat_avg_mol'], self.subject_weights)
        no['adj_int_avg_mol'] = self.adj_dilution_space(no['int_avg_mol'], self.subject_weights)
        no['adj_plat_avg_kg'] = no['adj_plat_avg_mol'] * STANDARD_WATER_MOL_MASS
        no['adj_int_avg_kg'] = no['adj_int_avg_mol'] * STANDARD_WATER_MOL_MASS
        return no

    @staticmethod
    def dilution_space_plateau(doseweight, molmass, doseenrichment, plateauenrichment, background):
        """Calculate a dilution space using the plateau method.
           :param doseweight: weight of the does in grams
           :param molmass: moleular mass of the dose in grams/mol
           :param doseenrichment: enrichment of the dose in ppm
           :param plateauenrichment: enrichment of the plateau urine measurement as a ratio
           :param background: enrichment of the background urine measurement as a ratio
           :return: dilution space calculated by the plateau method in mol
        """

        return doseweight / molmass * (doseenrichment - plateauenrichment) / (plateauenrichment - background)

    @staticmethod
    def dilution_space_intercept(doseweight, molmass, doseenrichment, plateauenrichment, background,
                                 turnoverrate, dosetime):
        """Calculate a dilution space using the intercept method.
           :param doseweight: weight of the does in grams
           :param molmass: moleular mass of the dose in grams/mol
           :param doseenrichment: enrichment of the dose in ppm
           :param plateauenrichment: enrichment of the plateau urine measurement in ppm
           :param background: enrichment of the background urine measurement in ppm
           :param turnoverrate: isotope turnover rate in 1/hr
           :param dosetime: time and date of dose administration
           :return: dilution space calculated by the intercept method in mol
        """

        interceptenrichment = (plateauenrichment - background) * np.exp(dosetime * turnoverrate) + background

        return doseweight / molmass * (doseenrichment - interceptenrichment) / (interceptenrichment - background)

    def avg_intercept_dilution_space(self, doseweight, molmass, doseenrichment, turnoverrate, ratios, sampledatetime):
        """Calculate the average of the 4 and 5 hour dilution spaces calculated by the intercept method
           :param doseweight: weight of the does in grams
           :param molmass: moleular mass of the dose in grams/mol
           :param doseenrichment: enrichment of the dose in ppm
           :param turnoverrate: isotope turnover rate in 1/hr
           :param ratios: measured urine isotope ratios
           :param sampledatetime: time and date of urine collections
           :return: average of the 4 and 5 hour dilution spaces calculated by the intercept method in mol
        """
        dilspace = np.zeros(2)
        dosetime = np.zeros(2)

        dosetime[0] = timedelta.total_seconds(sampledatetime[1] - sampledatetime[0]) / 3600
        dilspace[0] = self.dilution_space_intercept(doseweight, molmass, doseenrichment, ratios[1], ratios[0],
                                                    turnoverrate, dosetime[0])

        dosetime[1] = timedelta.total_seconds(sampledatetime[2] - sampledatetime[0]) / 3600
        dilspace[1] = self.dilution_space_intercept(doseweight, molmass, doseenrichment, ratios[2], ratios[0],
                                                    turnoverrate, dosetime[1])
        return np.mean(dilspace)

    @staticmethod
    def adj_dilution_space(dilution_space, subject_weights):
        """Adjust the given dilution space by the beginning and end subject weights to produce an average dilution space.
           :param dilution_space: unadjusted dilution space
           :param subject_weights: array containing before and after subject weights
           :return adj_dilution_space: dilution space in mol adjusted for the subject weight change over the
                                       sampling period
        """
        begin_dilution_spacekg = dilution_space * STANDARD_WATER_MOL_MASS
        end_dilution_spacekg = subject_weights[1] / subject_weights[0] * begin_dilution_spacekg

        adj_dilution_space = ((begin_dilution_spacekg + end_dilution_spacekg) / 2) / STANDARD_WATER_MOL_MASS
        return adj_dilution_space

    def calculate_schoeller(self):
        """Calculate the various rCO2 and TEE values using the equation of Schoeller (equation A6, 1986 as updated
            in 1988)
            :return dict of co2 and tee values from the Schoeller equation
        """
        schoeller = {}
        schoeller['co2_int'] = self.calc_schoeller_co2(self.nd['adj_int_avg_mol'], self.no['adj_int_avg_mol'],
                                                         self.kd_per_hr, self.ko_per_hr)
        schoeller['co2_plat'] = self.calc_schoeller_co2(self.nd['adj_plat_avg_mol'], self.no['adj_plat_avg_mol'],
                                                          self.kd_per_hr, self.ko_per_hr)
        schoeller = self.change_units_co2(schoeller)
        schoeller = self.tee_calcs(schoeller)

        return schoeller

    @staticmethod
    def calc_schoeller_co2(nd, no, kd, ko):
        """Calculate CO2 production in mol/hr using the equation of Schoeller (equation A6, 1986 as updated in 1988)
                from dilution spaces and isotope turnover rates.
           :param nd: deuterium dilution space in mol
           :param no: oxygen dilution space in mol
           :param kd: deuterium turnover rate in 1/hr
           :param ko: oxygen turnover rate in 1/hr
           :return co2prod: co2 production rate in mol/hr
        """
        n = ((no / 1.01) + (nd / 1.04)) / 2
        co2_prod = (n / 2.078) * (1.01 * ko - 1.04 * kd) - 0.0246 * n * 1.05 * (1.01 * ko - 1.04 * kd)
        # reduces to co2prod = n*(0.459952*ko_per_hr - 0.47362*kd_per_hr)
        return co2_prod

    def calculate_racette(self):
        """Calculate the various rCO2 and TEE values using the equation of Racette (1994)
            :return dict of co2 and tee values from the Racette equation
        """
        racette = {}
        racette['co2_int'] = self.calc_racette_co2(self.nd['adj_int_avg_mol'], self.no['adj_int_avg_mol'],
                                                         self.kd_per_hr, self.ko_per_hr, self.pop_avg_rdil)
        racette['co2_plat'] = self.calc_racette_co2(self.nd['adj_plat_avg_mol'], self.no['adj_plat_avg_mol'],
                                                          self.kd_per_hr, self.ko_per_hr, self.pop_avg_rdil)
        racette = self.change_units_co2(racette)
        racette = self.tee_calcs(racette)

        return racette

    @staticmethod
    def calc_racette_co2(nd, no, kd, ko, pop_avg_rdil):
        """Calculate CO2 production in mol/hr using the equation of Racette (1994)
                from dilution spaces and isotope turnover rates.
           :param nd: deuterium dilution space in mol
           :param no: oxygen dilution space in mol
           :param kd: deuterium turnover rate in 1/hr
           :param ko: oxygen turnover rate in 1/hr
           :return co2prod: co2 production rate in mol/hr
        """
        r_dil = (pop_avg_rdil+1.034)/2
        n = ((no / 1.01) + (nd / (1.01 *r_dil))) / 2
        co2_prod = (n / 2.078) * (1.01 * ko - 1.01 * kd * r_dil) - 0.0245 * n * 1.05 * (1.01 * ko - 1.01 * kd * r_dil)
        return co2_prod

    def calculate_speakman(self):
        """Calculate the various rCO2 and TEE values using the equation of Speakman eqn 17.41 (1997)
            :return dict of co2 and tee values from the Speakman equation
        """
        speakman = {}
        speakman['co2_int'] = self.calc_speakman_co2(self.nd['adj_int_avg_mol'], self.no['adj_int_avg_mol'],
                                                   self.kd_per_hr, self.ko_per_hr, self.pop_avg_rdil)
        speakman['co2_plat'] = self.calc_speakman_co2(self.nd['adj_plat_avg_mol'], self.no['adj_plat_avg_mol'],
                                                    self.kd_per_hr, self.ko_per_hr, self.pop_avg_rdil)
        speakman = self.change_units_co2(speakman)
        speakman = self.tee_calcs(speakman)

        return speakman

    @staticmethod
    def calc_speakman_co2(nd, no, kd, ko, pop_avg_rdil):
        """Calculate CO2 production in mol/hr using the equation of Speakman eqn 17.41 (1997)
                from dilution spaces and isotope turnover rates.
           :param nd: deuterium dilution space in mol
           :param no: oxygen dilution space in mol
           :param kd: deuterium turnover rate in 1/hr
           :param ko: oxygen turnover rate in 1/hr
           :return co2prod: co2 production rate in mol/hr
        """
        n = (no + (nd / pop_avg_rdil)) / 2
        co2_prod = (n / 2.078) * (ko - kd * pop_avg_rdil) - (0.0062 * n  * kd * pop_avg_rdil)
        return co2_prod

    @staticmethod
    def change_units_co2(equation):
        """Change the units on the co2 calculations.
            :param equation: dict with 'co2_int' and 'co2_plat' already calculated in mol/hr
            :return equation: dict now containing co2 in other units
        """

        equation['co2_int_mol_day'] = equation['co2_int'] * HOURS_PER_DAY  # rco2 mols/day
        equation['co2_int_L_hr'] = equation['co2_int'] * LITERS_PER_MOL
        equation['co2_int_L_day'] = equation['co2_int_L_hr'] * HOURS_PER_DAY  # rco2 l/day

        equation['co2_plat_mol_day'] = equation['co2_plat'] * HOURS_PER_DAY
        equation['co2_plat_L_hr'] = equation['co2_plat'] * LITERS_PER_MOL
        equation['co2_plat_L_day'] = equation['co2_plat_L_hr'] * HOURS_PER_DAY

        return equation

    @staticmethod
    def co2_to_tee(co2):
        """Convert CO2 production to total energy expenditure in using the equation of Weir, J.B. J Physiol., 109(1-2):1-9, 1949
           :param co2: volume of co2 production in mol/hr
           :return: total energy expenditure in kcal/day
        """
        return co2 * LITERS_PER_MOL * HOURS_PER_DAY * WEIR_CONSTANT

    def tee_calcs (self, equation):
        """Change the units on the tee calculations.
           :param equation: dict with co2 previously calculated
           :return equation: dict now containing tee measurements
        """

        equation['tee_int_kcal_day'] = self.co2_to_tee(equation['co2_int'])
        equation['tee_plat_kcal_day'] = self.co2_to_tee(equation['co2_plat'])

        equation['tee_int_mj_day'] = equation['tee_int_kcal_day'] * MJ_PER_KCAL
        equation['tee_plat_mj_day'] = equation['tee_plat_kcal_day'] * MJ_PER_KCAL

        return equation

    @staticmethod
    def percent_difference(first, second):
        """Calculate the percent difference between two values """
        return ((first - second) / ((first + second) / 2) * 100)

    def ee_consistency_check(self):
        """Calculate the percentage difference between the energy expenditure measured using the PD4/ED4 pair and
            the PD5/ED5 pair
            :return: percentage differences
        """
        elapsedhours = (timedelta.total_seconds(self.sample_datetimes[4] - self.sample_datetimes[2])) / 3600
        kd_a = self.isotope_turnover_2pt(self.d_ratios[0], self.d_ratios[1], self.d_ratios[3], elapsedhours)
        ko_a = self.isotope_turnover_2pt(self.o18_ratios[0], self.o18_ratios[1], self.o18_ratios[3], elapsedhours)

        elapsedhours = (timedelta.total_seconds(self.sample_datetimes[5] - self.sample_datetimes[3])) / 3600
        kd_b = self.isotope_turnover_2pt(self.d_ratios[0], self.d_ratios[2], self.d_ratios[4], elapsedhours)
        ko_b = self.isotope_turnover_2pt(self.o18_ratios[0], self.o18_ratios[2], self.o18_ratios[4], elapsedhours)

        schoeller_a = self.calc_schoeller_co2(self.nd['int_a_mol'], self.no['int_a_mol'], kd_a, ko_a)
        schoeller_b = self.calc_schoeller_co2(self.nd['int_b_mol'], self.no['int_b_mol'], kd_b, ko_b)

        tee_a = self.co2_to_tee(schoeller_a)
        tee_b = self.co2_to_tee(schoeller_b)

        diff = self.percent_difference(tee_a, tee_b)
        return diff

    def save_results_csv(self, filename):
        """ Save the results to a csv file
            :param: filename(string), the name of the file to which to save
        """
        write_header = ('subject_id,kd_hr,ko_hr,Nd_plat_avg_mol,No_plat_avg_mol, TBW_avg_kg,FFM_kg,FM_kg,body_fat_%,'
                        'sch_rCO2_int_mol/day,sch_rCO2_int_L/day,sch_EE_int_kcal/day,sch_EE_int_MJ/day,'
                        'sch_rCO2_plat_mol/day,sch_rCO2_plat_L/day,sch_EE_plat_kcal/day,sch_EE_plat_MJ/day,'
                        'rac_rCO2_int_mol/day,rac_rCO2_int_L/day,rac_EE_int_kcal/day,rac_EE_int_MJ/day,'
                        'rac_rCO2_plat_mol/day,rac_rCO2_plat_L/day,rac_EE_plat_kcal/day,rac_EE_plat_MJ/day,'
                        'spk_rCO2_int_mol/day,spk_rCO2_int_L/day,spk_EE_int_kcal/day,spk_EE_int_MJ/day,'
                        'spk_rCO2_plat_mol/day,spk_rCO2_plat_L/day,spk_EE_plat_kcal/day,spk_EE_plat_MJ/day,'
                        '2H_plateau_%,18O_plateau_%,DS_ratio,EE_consistency_check,ko/kd')
        write_data = np.asarray(
            [[self.subject_id, self.kd_per_hr, self.ko_per_hr, self.nd['plat_avg_mol'], self.no['plat_avg_mol'],
              self.total_body_water_ave_kg, self.fat_free_mass_kg, self.fat_mass_kg, self.body_fat_percent,
              self.schoeller['co2_int_mol_day'], self.schoeller['co2_int_L_day'], self.schoeller['tee_int_kcal_day'],
              self.schoeller['tee_int_mj_day'], self.schoeller['co2_plat_mol_day'], self.schoeller['co2_plat_L_hr'],
              self.schoeller['tee_plat_kcal_day'], self.schoeller['tee_plat_mj_day'],
              self.racette['co2_int_mol_day'], self.racette['co2_int_L_day'], self.racette['tee_int_kcal_day'],
              self.racette['tee_int_mj_day'], self.racette['co2_plat_mol_day'], self.racette['co2_plat_L_hr'],
              self.racette['tee_plat_kcal_day'], self.racette['tee_plat_mj_day'],
              self.speakman['co2_int_mol_day'], self.speakman['co2_int_L_day'], self.speakman['tee_int_kcal_day'],
              self.speakman['tee_int_mj_day'], self.speakman['co2_plat_mol_day'], self.speakman['co2_plat_L_hr'],
              self.speakman['tee_plat_kcal_day'], self.speakman['tee_plat_mj_day'], self.d_ratio_percent,
              self.o18_ratio_percent, self.dil_space_ratio, self.ee_check, self.ko_kd_ratio]])
        if os.path.isfile(filename):  # if the file already exists, don't rewrite the header
            file = open(filename, 'a+')
            np.savetxt(file, write_data, delimiter=',', comments='', fmt="%s")
        else:
            file = open(filename, 'a+')
            np.savetxt(file, write_data, delimiter=',', header=write_header, comments='', fmt="%s")
        file.close()
        return os.path.abspath(filename)



import setuptools

setuptools.setup(
        name='dlw',
        version='0.0.2',
        author='Julia Chmyz',
        author_email='julia@chmyz.com',
        packages=setuptools.find_packages(),
        include_package_data=True,
        url='https://github.com/jchmyz/DoublyLabeledWater',
        license='LICENSE.txt',
        description='Open source web app to conduct doubly labeled water study calculations.',
        long_description=open('README.md').read(),
        install_requires=[
            "numpy",
            "python-csv",
            "flask",
            "flask-restful"
        ],
        entry_points={"console_scripts": {"run-dlw = dlw.web.app:run_app"}}
)



from unittest import TestCase
import numpy as np
import datetime
import dlw
import os
import tempfile

D_DELTAS_TEST = np.array([-76.416, 772.106, 787.687, 242.117, 242.141])
O18_DELTAS_TEST = np.array([-7.984, 106.433, 108.619, 29.245, 29.104])
D_PPM_TEST = np.array([143.857444, 276.023231, 278.450127, 193.472144, 193.475882])
O18_PPM_TEST = np.array([1989.190483, 2218.619452, 2223.002819, 2063.842074, 2063.559341])

D_RATIOS_TEST = np.array([0.000143857, 0.000276023, 0.000278450, 0.000193472, 0.000193476])
O18_RATIOS_TEST = np.array([0.001989190, 0.002218619, 0.002223003, 0.002063842, 0.002063559])
BACKGROUND_TEST = datetime.datetime(2014, 2, 24, 6, 45)
DOSE_TEST = datetime.datetime(2014, 2, 24, 7, 00)  # year, month, day, hours, minutes
PD4_TEST = datetime.datetime(2014, 2, 24, 11, 00)
PD5_TEST = datetime.datetime(2014, 2, 24, 12, 00)
ED4_TEST = datetime.datetime(2014, 3, 3, 11, 00)
ED5_TEST = datetime.datetime(2014, 3, 3, 11, 58)
SAMPLE_DATETIME_TEST = [BACKGROUND_TEST, DOSE_TEST, PD4_TEST, PD5_TEST, ED4_TEST, ED5_TEST]
DOSE_WEIGHTS_TEST = [4.957, 8.649]
MIXED_DOSE_TEST = False
# MOL_MASSES_TEST = [19.9960, 19.9000]
DOSE_ENRICHMENTS_TEST = [998000, 950000]
SUBJECT_WEIGHTS_TEST = [59.62, 58.82]
SUBJECT_ID_TEST = "TestSubject"
POP_AVG_RDIL_TEST = 1.03

INCORRECT_RATIOS_TEST = np.array([-62.281, 742.928, 243.613, 739.377, 242.038])
INCORRECT_SIZE_TEST = np.array([-62.281, 742.928, 243.613, 739.377])
INCORRECT_DATES_TEST = [BACKGROUND_TEST, DOSE_TEST, PD4_TEST, PD5_TEST, PD4_TEST, ED5_TEST]


class TestDLWSubject(TestCase):
    """ Class for testing of DLWSubject class
    """

    def test_incorrect_sizes(self):
        with self.assertRaises(ValueError) as context:
            dlw.DLWSubject(INCORRECT_SIZE_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                           MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
            self.assertTrue('Arrays not correct size' in context.exception)

    def test_DLWsubject_inline_calcs(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(1.146923975, test_subject.ko_kd_ratio)
        self.assertAlmostEqual(1.04211133, test_subject.dil_space_ratio)
        self.assertAlmostEqual(32.02403252, test_subject.total_body_water_d_kg)
        self.assertAlmostEqual(31.75599874, test_subject.total_body_water_o_kg)
        self.assertAlmostEqual(31.89001563, test_subject.total_body_water_ave_kg)
        self.assertAlmostEqual(43.68495292, test_subject.fat_free_mass_kg)
        self.assertAlmostEqual(15.93504708, test_subject.fat_mass_kg)
        self.assertAlmostEqual(26.72768715, test_subject.body_fat_percent)

    def test_d_deltas_to_ratios(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertTrue(np.allclose(D_RATIOS_TEST, test_subject.d_ratios))

    def test_o18_deltas_to_ratios(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertTrue(np.allclose(O18_RATIOS_TEST, test_subject.o18_ratios))

    def test_d_ratios_to_deltas(self):
        test_subject = dlw.DLWSubject(D_PPM_TEST, O18_PPM_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST,
                                      False)
        self.assertTrue(np.allclose(D_DELTAS_TEST, test_subject.d_deltas))

    def test_o18_ratios_to_deltas(self):
        test_subject = dlw.DLWSubject(D_PPM_TEST, O18_PPM_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST,
                                      False)
        self.assertTrue(np.allclose(O18_DELTAS_TEST, test_subject.o18_deltas))

    def test_average_turnover_2pt(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(0.00588674101, test_subject.kd_per_hr)
        self.assertAlmostEqual(0.006751644, test_subject.ko_per_hr)

    def test_incorrect_ratios(self):
        with self.assertRaises(ValueError) as context:
            dlw.DLWSubject(INCORRECT_RATIOS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                           MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
            self.assertTrue(
                'Isotope ratios do not conform to pattern background < final < plateau' in context.exception)

    def test_calculate_mol_masses(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(20.0000104, test_subject.mol_masses[0])
        self.assertAlmostEqual(19.90031152, test_subject.mol_masses[1])

    def test_calculate_various_nd(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(1871.02702742, test_subject.nd['plat_a_mol'])
        self.assertAlmostEqual(1837.28529309, test_subject.nd['plat_b_mol'])
        self.assertAlmostEqual(1854.15616026, test_subject.nd['plat_avg_mol'])
        self.assertAlmostEqual(1827.47891022, test_subject.nd['int_a_mol'])
        self.assertAlmostEqual(1783.98809916, test_subject.nd['int_b_mol'])
        self.assertAlmostEqual(1805.73350469, test_subject.nd['int_avg_mol'])
        self.assertAlmostEqual(1841.71633362, test_subject.nd['adj_plat_avg_mol'])
        self.assertAlmostEqual(1793.61855330, test_subject.nd['adj_int_avg_mol'])
        self.assertAlmostEqual(33.3370178578, test_subject.nd['adj_plat_avg_kg'])
        self.assertAlmostEqual(32.4663970504, test_subject.nd['adj_int_avg_kg'])

    def test_calculate_various_no(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(1795.41951937, test_subject.no['plat_a_mol'])
        self.assertAlmostEqual(1761.75197101, test_subject.no['plat_b_mol'])
        self.assertAlmostEqual(1778.58574519, test_subject.no['plat_avg_mol'])
        self.assertAlmostEqual(1747.56869702, test_subject.no['int_a_mol'])
        self.assertAlmostEqual(1703.25659004, test_subject.no['int_b_mol'])
        self.assertAlmostEqual(1725.41264353, test_subject.no['int_avg_mol'])
        self.assertAlmostEqual(1766.65293241, test_subject.no['adj_plat_avg_mol'])
        self.assertAlmostEqual(1713.83657749, test_subject.no['adj_int_avg_mol'])
        self.assertAlmostEqual(31.9782907287, test_subject.no['adj_plat_avg_kg'])
        self.assertAlmostEqual(31.0222587193, test_subject.no['adj_int_avg_kg'])

    def test_calculate_schoeller(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(0.542979364, test_subject.schoeller['co2_int'])

    def test_calc_schoeller_co2(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(0.558617496, test_subject.schoeller['co2_plat'])

    def test_calculate_racette(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(0.556364985, test_subject.racette['co2_int'])

    def test_calc_racette_co2(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(0.572387168, test_subject.racette['co2_plat'])

    def test_calculate_speakman(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(0.559617317, test_subject.speakman['co2_int'])

    def test_pop_avg_rdil_incorporation(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST,
                                      pop_avg_rdil=POP_AVG_RDIL_TEST)
        self.assertAlmostEqual(0.50729412, test_subject.speakman['co2_int'])

    def test_calc_speakman_co2(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(0.575729332, test_subject.speakman['co2_plat'])

    def test_change_units_co2(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(13.03150474, test_subject.schoeller['co2_int_mol_day'])
        self.assertAlmostEqual(12.17033947, test_subject.schoeller['co2_int_L_hr'])
        self.assertAlmostEqual(292.0881473, test_subject.schoeller['co2_int_L_day'])
        self.assertAlmostEqual(13.4068199, test_subject.schoeller['co2_plat_mol_day'])
        self.assertAlmostEqual(12.52085255, test_subject.schoeller['co2_plat_L_hr'])
        self.assertAlmostEqual(300.5004613, test_subject.schoeller['co2_plat_L_day'])

    def test_tee_calcs(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(7.017890921, test_subject.schoeller['tee_int_mj_day'])
        self.assertAlmostEqual(7.220010393, test_subject.schoeller['tee_plat_mj_day'])

    def test_co2_to_tee(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(1725.62389885, test_subject.schoeller['tee_plat_kcal_day'])
        self.assertAlmostEqual(1677.31618574, test_subject.schoeller['tee_int_kcal_day'])

    def test_percent_difference(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(-1.8195460185388, test_subject.d_ratio_percent)

    def test_ee_consistency_check(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        self.assertAlmostEqual(-1.03567421899, test_subject.ee_check)

    def test_save_results_csv(self):
        test_subject = dlw.DLWSubject(D_DELTAS_TEST, O18_DELTAS_TEST, SAMPLE_DATETIME_TEST, DOSE_WEIGHTS_TEST,
                                      MIXED_DOSE_TEST, DOSE_ENRICHMENTS_TEST, SUBJECT_WEIGHTS_TEST, SUBJECT_ID_TEST)
        test_path = tempfile.NamedTemporaryFile().name
        test_subject.save_results_csv(test_path)
        read_data = np.genfromtxt(test_path, delimiter=',', skip_header=1, dtype=str)
        self.assertEqual('TestSubject', read_data[0])
        self.assertAlmostEqual(test_subject.no['plat_avg_mol'], float(read_data[4]))
        self.assertAlmostEqual(test_subject.body_fat_percent, float(read_data[8]))
        self.assertAlmostEqual(test_subject.schoeller['tee_int_kcal_day'], float(read_data[11]))
        self.assertAlmostEqual(test_subject.racette['tee_int_kcal_day'], float(read_data[19]))
        self.assertAlmostEqual(test_subject.speakman['tee_int_kcal_day'], float(read_data[27]))
        self.assertAlmostEqual(test_subject.ko_kd_ratio, float(read_data[37]))
